/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

/* Description: This program simply presents the usage of the statements of our API in succession.
 * (the output depends on the given mapping file).
 *
 * To compile this file use the Makefile located in the directory "examples"
 * and type the command "make api_test"
 *
 * This program requires the following:
 *
 *  - a globus user credentials, you can type:
 * 		grid-proxy-init -verify -debug
 *
 * 	- a machine file with the name "machines" in the same directory
 *    with the following content at least:
 *		"/your_hostname/" 9
 *
 * to execute this test using mpirun you can type e.g.:
 * 		mpirun -np 9 apiTest test.xml
 *
 * 			OR
 *
 *		../../bin/taagrun test.xml
 *
 * Remark: In the second case machine file is not required.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mpi.h>
#include <taag.h>

#define XML_DEFAULT "test.xml"

int main(int argc, char **argv) {
	int numprocs, my_id;
	int rc, flag, nrGrps, nrLANs, nrHosts, depth, root;
	int *dist;
	int i, j, k, l, content[100];
	char file_name[TAAG_MAX_FILENAME_STRING];
	char host_name[TAAG_MAX_HOSTNAME_STRING];
	double latency;


	rc = MPI_Init(&argc, &argv);
	if (rc != MPI_SUCCESS) {
		printf("Error starting MPI program.\n");
		MPI_Abort(MPI_COMM_WORLD, rc);
	}
	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &my_id);

	if (argc > 1) {
		strcpy(file_name, argv[1]);
	}
	else {
		strcpy(file_name, XML_DEFAULT);
	}

	if (my_id == 0) {
		printf("MPI environment is initialized with %d process(es).\n", numprocs);

		rc = TAAG_Initialized(&flag);
		if (rc != TAAG_SUCCESS)  {
			printf("Error in TAAG_Initialized() on process %d.\n",my_id);
			MPI_Finalize();
			return rc;
		}
		if (flag) {
			printf("- TAAG environment from xml source was initialized on process %d.\n", my_id);
		}
		else {
			printf("- TAAG environment was not initialized on process %d!\n", my_id);
		}
	}

	rc = TAAG_Init(file_name);
	if (rc != TAAG_SUCCESS) {
		printf("Error (code %d) initializing the TAAG structure on process %d.\n", rc, my_id);
		MPI_Finalize();
		return rc;
	}


	if (my_id == 0) {

		rc = TAAG_Initialized(&flag);
		if (rc != TAAG_SUCCESS) {
			printf("Error in TAAG_Initialized() on process %d.\n",my_id);
			MPI_Finalize();
			return rc;
		}
		if (flag) {
			printf("- TAAG environment from xml source was initialized on process %d.\n", my_id);
		}
		else {
			printf("- TAAG environment was not initialized on process %d!\n", my_id);
		}

		// ***** query the number of the groups
		rc = TAAG_Group_number(&nrGrps);
		if (rc != TAAG_SUCCESS) {
			printf("Error in TAAG_Group_number().\n");
			TAAG_Free();
			MPI_Finalize();
			return rc;
		}
		printf("- The number of groups is %d\n", nrGrps);

		// !!!From now I do not investigate the return values!!!

		// *****query the the id of a group that belongs to a particular process
		TAAG_Group_rank(my_id, &i);
		printf("- Process %d. belongs to the group %d.\n", my_id, i);

		// *****query the contents of the groups
		for (i = 0; i < nrGrps; i++) {
			TAAG_Group_size(i, &j);
			printf("- The size of the %d. group is %d\n", i, j);

			TAAG_Group_members(i, j, content);
			printf("  and it contains the following processes:");
			for (k = 0; k < j; k++){
				printf(" %d", content[k]);
			}
			printf("\n");
		}

		// ***** query the whole distance matrix of the groups
		dist = (int *)calloc(nrGrps * nrGrps, sizeof(int));
		TAAG_Group_graph(nrGrps, dist);
		printf("- The distance matrix is:\n");
		for (i = 0; i < nrGrps; i++) {
			for (j = 0; j < nrGrps; j++) {
				printf("\t%d",*(dist + i*nrGrps + j) );
			}
			printf("\n");
		}

		// ***** query the neigbourhood of a group
		TAAG_Group_degree(0, 1, &i);
		printf("- Group 0. has %d first-order neigbhor(s).\n",i);
		TAAG_Group_neighbours(0, 1, i, content);
		printf("  These groups are:");
		for (j = 0 ; j < i; j++){
			printf(" %d", content[j]);
		}
		printf("\n");

		TAAG_Group_degree(0, 2, &i);
		printf("- Group 0. has %d second-order neigbhor(s).\n",i);
		TAAG_Group_neighbours(0, 2, i, content);
		printf("  These groups are:");
		for (j = 0 ; j < i; j++){
			printf(" %d", content[j]);
		}
		printf("\n");

		TAAG_Group_degree(0, -1, &i);
		printf("- There are %d unreachable group(s) from group 0.\n",i);
		TAAG_Group_neighbours(0, -1, i, content);
		printf("  These groups are:");
		for (j = 0 ; j < i; j++) {
			printf(" %d", content[j]);
		}
		printf("\n");

		// ***** Checking the distance of two groups
		TAAG_Group_distance(0, nrGrps-1, &i);
		printf("- The distance between group 0 and group %d is %d steps.\n", nrGrps-1, i);
		TAAG_Group_way(0, nrGrps-1, i, content);
		printf("- The way from group 0 to group %d is:", nrGrps-1);
		for (j = 0; j < i; j++) {
			printf(" %d", content[j]);
		}
		printf("\n");

		// ***** query the rank of the last process in the last group
		TAAG_Group_size(nrGrps-1, &i);
		TAAG_Group_element(nrGrps-1, i-1, &j);
		printf("- The last process in the last group(%d.) is: %d\n", nrGrps-1,j);
		free(dist);

		// ***** query the hardware topology

		printf("**********Hardware Topology***********\n");
		TAAG_Topology_given(&flag);
		if (flag) {
			printf("- Network topology is described.\n");

			TAAG_LAN_number(&nrLANs);
			printf("- The number of LANs is %d.\n", nrLANs);
			TAAG_Host_number(&nrHosts);
			printf("- The number of hosts is %d.\n", nrHosts);
			TAAG_CPU_number(&i);
			printf("- The number of CPUs is %d.\n", i);

			TAAG_LAN_properties(my_id, &j, &k, &l);
			printf("\n- The LAN where the process %d runs consists of\n", my_id);
			printf("  \t- %d hosts,\n",j);
			printf("  \t- %d CPUs and\n", k);
			printf("  \t- executes %d processes.", l);
			TAAG_LAN_processes(my_id, l, content);
			printf("These processes are ");
			for (j = 0; j < l; j++) {
				printf(" %d", content[j]);
			}
			printf("\n");

			TAAG_LAN_properties(numprocs-1, &j, &k, &l);
			printf("\n- The LAN where the process %d runs consists of\n", numprocs-1);
			printf("  \t- %d hosts,\n",j);
			printf("  \t- %d CPUs and\n", k);
			printf("  \t- executes %d processes.", l);
			TAAG_LAN_processes(numprocs-1, l, content);
			printf("These processes are ");
			for (j = 0; j < l; j++) {
				printf(" %d", content[j]);
			}
			printf("\n");

			TAAG_Host_address(my_id, host_name);
			TAAG_Host_properties(my_id, &k, &l);
			TAAG_Host_processes(my_id, l, content);
			printf("\n- The host where the process %d runs called %s and\n", my_id, host_name);
			printf("  consists of\n");
			printf("  \t- %d CPUs and\n", k);
			printf("  \t- executes %d processes.", l);
			printf("These processes are ");
			for (j = 0; j < l; j++) {
				printf(" %d", content[j]);
			}
			printf("\n");

			TAAG_Host_address(numprocs-1, host_name);
			TAAG_Host_properties(numprocs-1, &k, &l);
			TAAG_Host_processes(numprocs-1, l, content);
			printf("\n- The host where the process %d runs called %s and\n", numprocs-1, host_name);
			printf("  consists of\n");
			printf("  \t- %d CPUs and\n", k);
			printf("  \t- executes %d processes.", l);
			printf("These processes are ");
			for (j = 0; j < l; j++) {
				printf(" %d", content[j]);
			}
			printf("\n");

			TAAG_Comm_level(my_id, numprocs-1, &k);
			printf("\n- The processes %d and %d can communicate each other on the level %d.\n", my_id, numprocs-1, k);
			printf("  Comment: WAN_LEVEL=%d LAN_LEVEL=%d HOST_LEVEL=%d.\n", TAAG_WAN_LEVEL, TAAG_LAN_LEVEL, TAAG_HOST_LEVEL);

			TAAG_Host_latency(my_id, numprocs-1, &latency);
			printf("- The latency value between the hosts where processes %d and %d reside is %.4f.\n", my_id, numprocs-1, latency);
		}
		else {
			printf("- Network topology is NOT described!\n");
		}
		printf("**************************************\n");

		TAAG_Tree_isTree(&flag);
		if (flag) {
			printf("- It is a tree structure.\n");

			TAAG_Tree_depth(&depth);
			printf("- The depth of the tree is %d.\n", depth);

			TAAG_Tree_root(&root);
			printf("- The rank of the root process is %d.\n", root);

			TAAG_Tree_level(root, &i);
			printf("- The level of process %d in the tree is %d.\n", root, i);
			TAAG_Tree_level(numprocs-1, &i);
			printf("- The level of process %d in the tree is %d.\n", numprocs-1, i);

			for (i = 0 ; i < depth; i++) {
				TAAG_Tree_width(root, i, &l);
				printf("- The width of the tree (in terms of processes) on level %d is %d.\n", i, l);
				TAAG_Tree_children(root, i, l, content);
				printf("- The %d. order children of the root are", i);
				for (j = 0; j < l; j++) {
					printf(" %d",content[j]);
				}
				printf("\n");
			}

			TAAG_Tree_parent(numprocs-1, &i);
			printf("- The parent of process %d is process %d.\n", numprocs-1, i);

			for (i=0 ; i < numprocs; i++) {
				TAAG_Tree_isLeaf(i, &flag);
				if (flag) {
					printf("-The process %d is a leaf.\n",i);
				}
				else {
					TAAG_Tree_isLocalManager(i, &flag);
					if (flag) {
						printf("-The process %d is a local manager.\n",i);
					}
					else {
						printf("-The process %d is neither a leaf nor a local manager.\n", i);
					}
				}
			} //for

		}
		else {
			printf("- It is NOT a tree structure!\n");
		}

	} //if

	rc = TAAG_Free();
	if (rc != TAAG_SUCCESS) {
		printf("Error in setting free the TAAG structure on process %d.\n\n", my_id);
	}

	MPI_Finalize();

	return 0;
}
