/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2010 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include <taag_basic.h>
#include <taag_errno.h>
#include <taag_schema_xml.h>

/*
 *To compile this file using gcc you can type
 *gcc `xml2-config --cflags --libs` -o ouput filenames
 */

static int xmlFileOpened = TAAG_FALSE;
static xmlDocPtr doc = NULL;
static xmlNode *root = NULL;

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_openFile(char *filename, char *type) {
	// Open XML document
	doc = xmlParseFile(filename);
	if (doc == NULL) return(TAAG_ERR_XML);

	//Get and check the root element node
	root = xmlDocGetRootElement(doc);
	if(!root || !root->name || xmlStrcmp(root->name, (const xmlChar *) TAAG_SCHEMA_ROOT) )
	{
		xmlFreeDoc(doc);
		return (TAAG_ERR_SCHEMA);
	} //if

	xmlChar *attr;
	attr = xmlGetProp(root, (const xmlChar *) TAAG_SCHEMA_ROOT_ATTR);
	if (attr== NULL) {
		xmlFreeDoc(doc);
		return (TAAG_ERR_SCHEMA);
	} //if
	strcpy(type, (char *)attr);

	xmlFileOpened = TAAG_TRUE;
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_fileOpened(int *flag) {
	*flag = xmlFileOpened;
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_validate() {
	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);
	//???
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_free() {
	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);
	/*free the document */
	xmlFreeDoc(doc);

	/*
	 *Free the global variables that may
	 *have been allocated by the parser.
	 */
	xmlCleanupParser();
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getType(char *type) {
	xmlChar *attr;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	attr = xmlGetProp(root, (const xmlChar *) TAAG_SCHEMA_ROOT_ATTR);
	strcpy(type, (char *)attr);
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getExecutable(char *executable) {
	xmlNode *node;
	xmlChar *fieldContent;
	int found = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (node = root->children; node != NULL && !found; node = node->next)
	{
		if (node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_APP_NAME) )
		{
			fieldContent = xmlNodeGetContent(node);
			strcpy(executable, (char *)fieldContent);
			found = TAAG_TRUE;
			xmlFree(fieldContent);
	     } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getDirectory(char *dir) {
	xmlNode *node;
	xmlChar *fieldContent;
	int found = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (node = root->children; node != NULL && !found; node = node->next)
	{
		if (node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_DIRECTORY) )
		{
			fieldContent = xmlNodeGetContent(node);
			strcpy(dir, (char *)fieldContent);
			found = TAAG_TRUE;
			xmlFree(fieldContent);
	     } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getGroups(int *nr, int *minSize, int *div) {
	xmlNode *node;
	xmlChar *fieldContent;
	int found_nr = TAAG_FALSE;
	int found_minSize = TAAG_FALSE;
	int found_div = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (node = root->children; node != NULL; node = node->next)
	{
		if (node->type == XML_ELEMENT_NODE) {
			if (!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_NR_PROCESSES) ) {
				fieldContent = xmlNodeGetContent(node);
				*nr = atoi((char *)fieldContent);
				found_nr = TAAG_TRUE;
				xmlFree(fieldContent);
			} //if
			else if (!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_MIN_SIZE_GRPS) ) {
				fieldContent = xmlNodeGetContent(node);
				*minSize = atoi((char *)fieldContent);
				found_minSize = TAAG_TRUE;
				xmlFree(fieldContent);
			} // else if
			else if (!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_DIVISOR) ) {
				fieldContent = xmlNodeGetContent(node);
				*div = atoi((char *)fieldContent);
				found_div = TAAG_TRUE;
				xmlFree(fieldContent);
			} //else if
		 } //if
	} //for

	if (found_nr && found_minSize && found_div) {
		return(TAAG_SUCCESS);
	} //if
	else {
		return(TAAG_ERR_SCHEMA);
	}
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getTree(int *nr, int *level, int *minSize) {
	xmlNode *node;
	xmlChar *fieldContent;
	int found_nr = TAAG_FALSE;
	int found_level = TAAG_FALSE;
	int found_minSize = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (node = root->children; node != NULL; node = node->next)
	{
		if (node->type == XML_ELEMENT_NODE) {
			if (!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_NR_PROCESSES) ) {
				fieldContent = xmlNodeGetContent(node);
				*nr = atoi((char *)fieldContent);
				found_nr = TAAG_TRUE;
				xmlFree(fieldContent);
			} //if
			else if (!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_MIN_SIZE_GRPS) ) {
				fieldContent = xmlNodeGetContent(node);
				*minSize = atoi((char *)fieldContent);
				found_minSize = TAAG_TRUE;
				xmlFree(fieldContent);
			} //else if
			else if (!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_NR_LEVELS) ) {
				fieldContent = xmlNodeGetContent(node);
				*level = atoi((char *)fieldContent);
				found_level = TAAG_TRUE;
				xmlFree(fieldContent);
			} //else if
		} //if
	} //for

	if (found_nr && found_minSize && found_level) {
		return(TAAG_SUCCESS);
	} //if
	else {
		return(TAAG_ERR_SCHEMA);
	}
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getRing(int *nr, int *minSize, int *div) {
	return _taag_schema_getGroups(nr, minSize, div);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getNumberOfGroups(int *nrOfGrps) {
	xmlNode *node;
	int grp_counter;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	grp_counter = 0;
	for(node = root->children; node != NULL; node = node->next)
	{
		if (node->type == XML_ELEMENT_NODE
				&&
			!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_SIZE_GRP) )
		{
			grp_counter++;
		} //if
	}//for
	*nrOfGrps = grp_counter;

	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getGroupSizes(int nrOfGrps, int *sizesOfGroups) {
	xmlNode *node;
	int grp_counter;
	xmlChar *fieldContent;
	xmlChar *attr;
	int returnValue = TAAG_SUCCESS;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	grp_counter = 0;
	for (node = root->children; node != NULL && grp_counter < nrOfGrps; node = node->next)
	{
		if (node->type == XML_ELEMENT_NODE
				&&
			!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_SIZE_GRP) )
		{
			attr = xmlGetProp(node, (const xmlChar *) TAAG_SCHEMA_ATTR_GRP_ID);
			if (grp_counter != atoi((char *)attr)) {
				returnValue = TAAG_ERR_XML;
			}

			fieldContent = xmlNodeGetContent(node);
			*(sizesOfGroups + grp_counter) = atoi((char *)fieldContent);
			xmlFree(fieldContent);
			grp_counter++;
		} //if
	}//for
	return(returnValue);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_schema_getEdgeMatrix(int nrOfGrps, int *edge_counter, int *edges) {
	xmlNode *node;
	xmlChar *fieldProp1, *fieldProp2;
	int grp1, grp2;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	*edge_counter = 0;
	for (node = root->children; node != NULL; node = node->next)
	{
		if (node->type == XML_ELEMENT_NODE
				&&
			!xmlStrcmp(node->name, (const xmlChar *) TAAG_SCHEMA_EDGE) )
		{
			fieldProp1 = xmlGetProp(node, (const xmlChar *) TAAG_SCHEMA_EDGE_ATTR_1);
			fieldProp2 = xmlGetProp(node, (const xmlChar *) TAAG_SCHEMA_EDGE_ATTR_2);
			grp1 = atoi((char *)fieldProp1);
			grp2 = atoi((char *)fieldProp2);
			if (grp1 < 0 || grp2 < 0 || grp1 >= nrOfGrps || grp2 >= nrOfGrps) return(TAAG_ERR_XML);
			*(edges+grp1*nrOfGrps+grp2) = 1;
			*(edges+grp2*nrOfGrps+grp1) = 1;
			(*edge_counter)++;
			xmlFree(fieldProp1);
			xmlFree(fieldProp2);
		} //if
	}//for
	return(TAAG_SUCCESS);
}
