/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include <taag_basic.h>
#include <taag_errno.h>
#include <taag_xml.h>

/*
 *To compile this file using gcc you can type
 *gcc `xml2-config --cflags --libs` -o ouput filenames
 */

static int xmlFileOpened = TAAG_FALSE;
static xmlDocPtr doc = NULL;
static xmlNode *root = NULL;

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_openFile(char *filename) {
	// Open XML document
	doc = xmlParseFile(filename);
	if (doc == NULL) return(TAAG_ERR_XML);

	//Get and check the root element node
	root = xmlDocGetRootElement(doc);
	if(!root || !root->name || xmlStrcmp(root->name, (const xmlChar *) TAAG_XML_ROOT) )
	{
		xmlFreeDoc(doc);
		return (TAAG_ERR_XML);
	} //if

	xmlFileOpened = TAAG_TRUE;
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_fileOpened(int *flag) {
	*flag = xmlFileOpened;
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_validate() {
	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);
	//???
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_free() {
	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);
	/*free the document */
	xmlFreeDoc(doc);

	/*
	 *Free the global variables that may
	 *have been allocated by the parser.
	 */
	xmlCleanupParser();
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getExecutable(char *executable) {
	xmlNode *exec_node;
	xmlChar *fieldContent;
	int found = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (exec_node = root->children; exec_node != NULL && !found; exec_node = exec_node->next)
	{
		if (exec_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(exec_node->name, (const xmlChar *) TAAG_XML_APP_NAME) )
		{
			fieldContent = xmlNodeGetContent(exec_node);
			strcpy(executable, (char *)fieldContent);
			found = TAAG_TRUE;
			xmlFree(fieldContent);
	     } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getAppId(char *applId) {
	xmlNode *curr_node;
	xmlChar *fieldContent;
	int found = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (curr_node = root->children; curr_node != NULL && !found; curr_node = curr_node->next)
	{
		if (curr_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(curr_node->name, (const xmlChar *) TAAG_XML_APP_ID) )
		{
			fieldContent = xmlNodeGetContent(curr_node);
			strcpy(applId, (char *)fieldContent);
			found = TAAG_TRUE;
			xmlFree(fieldContent);
	     } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getTimeStamp(char *timeStamp) {
	xmlNode *curr_node;
	xmlChar *fieldContent;
	int found = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (curr_node = root->children; curr_node != NULL && !found; curr_node = curr_node->next)
	{
		if (curr_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(curr_node->name, (const xmlChar *) TAAG_XML_TIME_STAMP) )
		{
			fieldContent = xmlNodeGetContent(curr_node);
			strcpy(timeStamp, (char *)fieldContent);
			found = TAAG_TRUE;
			xmlFree(fieldContent);
	     } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getSchemaType(char *type) {
	xmlNode *graph_node, *cur_node;
	xmlChar *fieldContent;
	int found = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (graph_node = root->children; graph_node != NULL && !found; graph_node = graph_node->next)
	{
		if (graph_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(graph_node->name, (const xmlChar *) TAAG_XML_GRAPH) )
		{
			for(cur_node = graph_node->children; cur_node != NULL && !found; cur_node = cur_node->next)
			{
				if (cur_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_GRAPH_TYPE) )
	    	    {
					fieldContent = xmlNodeGetContent(cur_node);
					strcpy(type, (char *)fieldContent);
					found = TAAG_TRUE;
					xmlFree(fieldContent);
	    	    } //if
			} //for
	     } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getNumberOfGroups(unsigned int *nr) {
	xmlNode *graph_node, *cur_node;
	int found = TAAG_FALSE;
	int grp_counter;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (graph_node = root->children; graph_node != NULL && !found; graph_node = graph_node->next)
	{
		if (graph_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(graph_node->name, (const xmlChar *) TAAG_XML_GRAPH) )
		{
			found = TAAG_TRUE;
			grp_counter = 0;
			for(cur_node = graph_node->children; cur_node != NULL; cur_node = cur_node->next)
			{
				if (cur_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_GROUP) )
		    	{
					grp_counter++;
		    	} //if
			}//for
			*nr = grp_counter;
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getGroupSizes(unsigned int nr, unsigned int *size) {
	xmlNode *graph_node, *group_node, *cur_node;
	int found = TAAG_FALSE;
	int grp_counter, procs_counter;
	xmlChar *attr;
	int returnValue = TAAG_SUCCESS;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (graph_node = root->children; graph_node != NULL && !found; graph_node = graph_node->next)
	{
		if (graph_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(graph_node->name, (const xmlChar *) TAAG_XML_GRAPH) )
		{
			found = TAAG_TRUE;
			grp_counter = 0;
			for (group_node = graph_node->children; group_node != NULL && grp_counter < nr; group_node = group_node->next)
			{
				if (group_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(group_node->name, (const xmlChar *) TAAG_XML_GROUP) )
		    	{
					attr = xmlGetProp(group_node, (const xmlChar *) TAAG_XML_ATTR_ID);
					if (grp_counter != atoi((char *)attr)) {
						returnValue = TAAG_ERR_XML;
					}

					procs_counter = 0;
					for (cur_node = group_node->children; cur_node != NULL; cur_node = cur_node->next)
					{
						if (cur_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_G_PROCS) )
						{
							procs_counter++;
						} //if
					} //for
					*(size+grp_counter) = procs_counter;
					grp_counter++;
		    	} //if
			}//for
		 } //if
	} //for
	return(returnValue);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getGroupMembers(unsigned int nrOfG, unsigned int *nrOfP, unsigned int **members) {
	xmlNode *graph_node, *group_node, *cur_node;
	xmlChar *fieldContent;
	int found = TAAG_FALSE;
	int grp_counter, procs_counter;
	xmlChar *attr;
	int returnValue = TAAG_SUCCESS;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (graph_node = root->children; graph_node != NULL && !found; graph_node = graph_node->next)
	{
		if (graph_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(graph_node->name, (const xmlChar *) TAAG_XML_GRAPH) )
		{
			found = TAAG_TRUE;
			grp_counter = 0;
			for (group_node = graph_node->children; group_node != NULL && grp_counter < nrOfG; group_node = group_node->next)
			{
				if (group_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(group_node->name, (const xmlChar *) TAAG_XML_GROUP) )
		    	{
					attr = xmlGetProp(group_node, (const xmlChar *) TAAG_XML_ATTR_ID);
					if (grp_counter != atoi((char *)attr)) {
						returnValue = TAAG_ERR_XML;
					}

					procs_counter = 0;
					for (cur_node = group_node->children; cur_node != NULL && procs_counter < *(nrOfP + grp_counter); cur_node = cur_node->next)
					{
						if (cur_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_G_PROCS) )
						{
							fieldContent = xmlNodeGetContent(cur_node);
							*(*(members + grp_counter) + procs_counter) = atoi((char *)fieldContent);
							xmlFree(fieldContent);
							procs_counter++;
						} //if
					} //for
					grp_counter++;
		    	} //if
			}//for
		 } //if
	} //for
	return(returnValue);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getEdgeMatrix(unsigned int nr, int *distances) {
	xmlNode *graph_node, *edge_node;
	int found = TAAG_FALSE;
	xmlChar *fieldProp1, *fieldProp2;
	int grp1, grp2;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (graph_node = root->children; graph_node != NULL && !found; graph_node = graph_node->next)
	{
		if (graph_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(graph_node->name, (const xmlChar *) TAAG_XML_GRAPH) )
		{
			found = TAAG_TRUE;

			for (edge_node = graph_node->children; edge_node != NULL; edge_node = edge_node->next)
			{
				if (edge_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(edge_node->name, (const xmlChar *) TAAG_XML_EDGE) )
		    	{
					fieldProp1 = xmlGetProp(edge_node, (const xmlChar *) TAAG_XML_ATTR_FIRST_END);
					fieldProp2 = xmlGetProp(edge_node, (const xmlChar *) TAAG_XML_ATTR_OTHER_END);
					grp1 = atoi((char *)fieldProp1);
					grp2 = atoi((char *)fieldProp2);
					if (grp1 < 0 || grp2 < 0 || grp1 >= nr || grp2 >= nr) return(TAAG_ERR_XML);
					*(distances+grp1*nr+grp2) = 1;
					*(distances+grp2*nr+grp1) = 1;
					xmlFree(fieldProp1);
					xmlFree(fieldProp2);
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}


/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_isTopologyGiven(int *flag) {
	xmlNode *topo_node;
	*flag = TAAG_FALSE;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !(*flag); topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			*flag = TAAG_TRUE;
		} //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getNumberOfLANs(unsigned int *nr) {
	xmlNode *topo_node, *cur_node;
	int found = TAAG_FALSE;
	int lan_counter;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;
			lan_counter = 0;
			for(cur_node = topo_node->children; cur_node != NULL; cur_node = cur_node->next)
			{
				if (cur_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_LAN) )
		    	{
					lan_counter++;
		    	} //if
				else if (cur_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_HOST) )
		    	{
					fprintf(stderr,"The XML tag \"%s\" always must be embedded into the XML  tag \"%s\".\n", TAAG_XML_HOST, TAAG_XML_LAN);
					return(TAAG_ERR_XML);
		    	} //if
			}//for
			*nr = lan_counter;
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getHosts(unsigned int nrLAN, int *host) {
	xmlNode *topo_node, *lan_node, *cur_node;
	int found = TAAG_FALSE;
	int lan_counter;
	int host_counter = 0;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;
			lan_counter = 0;
			for (lan_node = topo_node->children; lan_node != NULL && lan_counter < nrLAN; lan_node = lan_node->next)
			{
				if (lan_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(lan_node->name, (const xmlChar *) TAAG_XML_LAN) )
		    	{
					for (cur_node = lan_node->children; cur_node != NULL; cur_node = cur_node->next)
					{
						if (cur_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_HOST) )
						{
							host_counter++;
						} //if
					} //for
					*(host+lan_counter) = host_counter;
					lan_counter++;
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 * Returns all the hostnames
 ***********************************************************************/
int _taag_xml_getHostnames(int nrHost, char **hostname) {
	xmlNode *topo_node, *lan_node, *host_node, *cur_node;
	int found = TAAG_FALSE;
	int name_found = TAAG_FALSE;
	int host_counter = 0;
	xmlChar *fieldContent;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;
			for (lan_node = topo_node->children; lan_node != NULL; lan_node = lan_node->next)
			{
				if (lan_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(lan_node->name, (const xmlChar *) TAAG_XML_LAN) )
		    	{
					for (host_node = lan_node->children; host_node != NULL && host_counter < nrHost; host_node = host_node->next)
					{
						if (host_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(host_node->name, (const xmlChar *) TAAG_XML_HOST) )
						{
							for (cur_node = host_node->children; cur_node != NULL && !name_found; cur_node = cur_node->next)
							{
								if (cur_node->type == XML_ELEMENT_NODE
										&&
									!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_H_NAME) )
								{
									name_found = TAAG_TRUE;
									fieldContent = xmlNodeGetContent(cur_node);
									strcpy(*(hostname+host_counter), (char *)fieldContent);
									xmlFree(fieldContent);
								} //if
							} //for
							host_counter++;
							name_found = TAAG_FALSE;
						} //if
					} //for
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getNumberOfCPUs(int nrHost, int *nrCPUs) {
	xmlNode *topo_node, *lan_node, *host_node, *cur_node;
	int found = TAAG_FALSE;
	int cpuInfo_found = TAAG_FALSE;
	int host_counter = 0;
	xmlChar *fieldContent;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;
			for (lan_node = topo_node->children; lan_node != NULL; lan_node = lan_node->next)
			{
				if (lan_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(lan_node->name, (const xmlChar *) TAAG_XML_LAN) )
		    	{
					for (host_node = lan_node->children; host_node != NULL && host_counter < nrHost; host_node = host_node->next)
					{
						if (host_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(host_node->name, (const xmlChar *) TAAG_XML_HOST) )
						{
							for (cur_node = host_node->children; cur_node != NULL && !cpuInfo_found; cur_node = cur_node->next)
							{
								if (cur_node->type == XML_ELEMENT_NODE
										&&
									!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_H_CPUS) )
								{
									cpuInfo_found = TAAG_TRUE;
									fieldContent = xmlNodeGetContent(cur_node);
									*(nrCPUs+host_counter) = atoi((char *)fieldContent);;
									xmlFree(fieldContent);
								} //if
							} //for
							host_counter++;
							cpuInfo_found = TAAG_FALSE;
						} //if
					} //for
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 * Returns all the application directories
 ***********************************************************************/
int _taag_xml_getDirectories(int nrHost, char **directories) {
	xmlNode *topo_node, *lan_node, *host_node, *cur_node;
	int found = TAAG_FALSE;
	int dir_found = TAAG_FALSE;
	int host_counter = 0;
	xmlChar *fieldContent;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;
			for (lan_node = topo_node->children; lan_node != NULL; lan_node = lan_node->next)
			{
				if (lan_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(lan_node->name, (const xmlChar *) TAAG_XML_LAN) )
		    	{
					for (host_node = lan_node->children; host_node != NULL && host_counter < nrHost; host_node = host_node->next)
					{
						if (host_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(host_node->name, (const xmlChar *) TAAG_XML_HOST) )
						{
							for (cur_node = host_node->children; cur_node != NULL && !dir_found; cur_node = cur_node->next)
							{
								if (cur_node->type == XML_ELEMENT_NODE
										&&
									!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_DIRECTORY) )
								{
									dir_found = TAAG_TRUE;
									fieldContent = xmlNodeGetContent(cur_node);
									strcpy(*(directories+host_counter), (char *)fieldContent);
									xmlFree(fieldContent);
								} //if
							} //for
							host_counter++;
							dir_found = TAAG_FALSE;
						} //if
					} //for
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getLatency(int nrHost, double *latency) {
	xmlNode *topo_node, *latency_node;
	int found = TAAG_FALSE;
	xmlChar *fieldProp1, *fieldProp2;
	xmlChar *fieldContent;
	int host1, host2;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;

			for (latency_node = topo_node->children; latency_node != NULL; latency_node = latency_node->next)
			{
				if (latency_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(latency_node->name, (const xmlChar *) TAAG_XML_LATENCY) )
		    	{

					fieldProp1 = xmlGetProp(latency_node, (const xmlChar *) TAAG_XML_ATTR_ROW);
					fieldProp2 = xmlGetProp(latency_node, (const xmlChar *) TAAG_XML_ATTR_COLUMN);
					host1 = atoi((char *)fieldProp1);
					host2 = atoi((char *)fieldProp2);
					if (host1 < 0 || host2 < 0 || host1 >= nrHost || host2 >= nrHost) return(TAAG_ERR_XML);
					fieldContent = xmlNodeGetContent(latency_node);
					*(latency+host1*nrHost+host2) = atof((char *)fieldContent);
					*(latency+host2*nrHost+host1) = atof((char *)fieldContent);
					xmlFree(fieldProp1);
					xmlFree(fieldProp2);
					xmlFree(fieldContent);
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getNumberOfProcsOnHosts(int nrHost, int *nrProcs) {
	xmlNode *topo_node, *lan_node, *host_node, *cur_node;
	int found = TAAG_FALSE;
	int host_counter = 0;
	int procs_counter;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;
			for (lan_node = topo_node->children; lan_node != NULL; lan_node = lan_node->next)
			{
				if (lan_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(lan_node->name, (const xmlChar *) TAAG_XML_LAN) )
		    	{
					for (host_node = lan_node->children; host_node != NULL && host_counter < nrHost; host_node = host_node->next)
					{
						if (host_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(host_node->name, (const xmlChar *) TAAG_XML_HOST) )
						{
							procs_counter = 0;
							for (cur_node = host_node->children; cur_node != NULL; cur_node = cur_node->next)
								{
								if (cur_node->type == XML_ELEMENT_NODE
										&&
									!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_H_PROCS) )
								{
									procs_counter++;
								} //if
							} //for
							*(nrProcs+host_counter) = procs_counter;
							host_counter++;
						} //if
					} //for
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}

/***********************************************************************
 *
 ***********************************************************************/
int _taag_xml_getProcsOnHosts(int nrHost, int *nrProcs, int **procs) {
	xmlNode *topo_node, *lan_node, *host_node, *cur_node;
	int found = TAAG_FALSE;
	int host_counter = 0;
	int procs_counter;
	xmlChar *fieldContent;

	if (xmlFileOpened == TAAG_FALSE) return(TAAG_ERR_XML);

	for (topo_node = root->children; topo_node != NULL && !found; topo_node = topo_node->next)
	{
		if (topo_node->type == XML_ELEMENT_NODE
	    		 &&
	    	!xmlStrcmp(topo_node->name, (const xmlChar *) TAAG_XML_TOPOLOGY) )
		{
			found = TAAG_TRUE;
			for (lan_node = topo_node->children; lan_node != NULL; lan_node = lan_node->next)
			{
				if (lan_node->type == XML_ELEMENT_NODE
						&&
					!xmlStrcmp(lan_node->name, (const xmlChar *) TAAG_XML_LAN) )
		    	{
					for (host_node = lan_node->children; host_node != NULL && host_counter < nrHost; host_node = host_node->next)
					{
						if (host_node->type == XML_ELEMENT_NODE
								&&
							!xmlStrcmp(host_node->name, (const xmlChar *) TAAG_XML_HOST) )
						{
							procs_counter = 0;
							for (cur_node = host_node->children; cur_node != NULL; cur_node = cur_node->next)
								{
								if (cur_node->type == XML_ELEMENT_NODE
										&&
									!xmlStrcmp(cur_node->name, (const xmlChar *) TAAG_XML_H_PROCS) )
								{
									fieldContent = xmlNodeGetContent(cur_node);
									*(*(procs + host_counter) + procs_counter) = atoi((char *)fieldContent);
									xmlFree(fieldContent);
									procs_counter++;
								} //if
							} //for
							host_counter++;
						} //if
					} //for
		    	} //if
			}//for
		 } //if
	} //for
	return(TAAG_SUCCESS);
}
