/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2010 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include "ProcsDistribution.h"

ProcsDistribution::ProcsDistribution() {
	mapped = false;
	edges = NULL;
}

ProcsDistribution::~ProcsDistribution() {
	if (edges!=NULL) delete[] edges;
	while (!grpList.empty()) {
		GroupData *gd = grpList.back();
		grpList.pop_back();
		delete gd;
	}
}

int ProcsDistribution::addGroup(GroupData *newGrp) {
	grpList.push_back(newGrp);
	return grpList.size();
}

int ProcsDistribution::getNrOfGroups() {
	return grpList.size();
}

GroupData *ProcsDistribution::getGroupByIndex(unsigned int index) {
	if (grpList.empty() || index >= grpList.size() || index < 0) return NULL;
	return grpList[index];
}

GroupData *ProcsDistribution::getLastGroup() {
	if (grpList.empty()) return NULL;
	return grpList.back();
}

void ProcsDistribution::removeLastGroup() {
	if (grpList.empty()) return;
	grpList.pop_back();
}

//************* EDGES begin********************
void ProcsDistribution::initEdgeMatrix() {
	int size = getNrOfGroups() * getNrOfGroups();
	edges = new bool[size];
	for (int i=0; i<size; i++) {
		edges[i] = false;
	}
}

void ProcsDistribution::setEdge(int grp1, int grp2) {
	int queueLength = getNrOfGroups();
	if (grp1<0 || grp2<0 || grp1>=queueLength || grp2>=queueLength) return;
	edges[grp1*queueLength+grp2] = true;
	edges[grp2*queueLength+grp1] = true;
	//printf("%d ---> %d (%d)\n", grp1, grp2, grp1*queueLength+grp2);
}

bool ProcsDistribution::getEdge(int grp1, int grp2) {
	int queueLength = getNrOfGroups();
	if (grp1<0 || grp2<0 || grp1>=queueLength || grp2>=queueLength) return 0;
	return edges[grp1*queueLength+grp2];
}

bool ProcsDistribution::areEdgesDefined() {
	return (edges!=NULL);
}
//************* EDGES end  ********************


void ProcsDistribution::calculateGreatestGroupLatencyLevel() {
	int greatest = 0;
	vector<GroupData *>::iterator i;
	for (i = grpList.begin(); i != grpList.end(); i++) {
		if ( (*i)->getTmpLevel() > greatest) greatest = (*i)->getTmpLevel();
	}
	greatestGroupTmpLevel = greatest;
}

double ProcsDistribution::mappingQuality() {
	if (!mapped || overallAvgOfLatencies == 0.0) return 0.0;
	return greatestGroupLatencyLevel+maxLatencyAmongGrps/overallAvgOfLatencies; // the second is divided by overall avg latance value
	//return greatestGroupLatencyLevel+(maxLatencyAmongGrps+avgLatencyAmongGrps)/overallAvgOfLatencies; //alternate quality function
}

double ProcsDistribution::tmpQuality() {
	if (overallAvgOfLatencies == 0.0) return 0.0;
	return greatestGroupTmpLevel+maxTmpAmongGrps/overallAvgOfLatencies; // the second is divided by overall avg latance value
	//return greatestGroupTmpLevel+(maxTmpAmongGrps+avgTmpAmongGrps)/overallAvgOfLatencies; //alternate quality function
}

void ProcsDistribution::setMapped() {
	mapped = true;
}
bool ProcsDistribution::isMapped() {
	return mapped;
}

void ProcsDistribution::printOutDistribution(FILE *f) {
	if (grpList.empty()) return;
	vector<GroupData *>::iterator it;
	for (it = grpList.begin(); it != grpList.end(); it++) {
		fprintf(f, "\t[%d]", (*it)->getNrOfProcs());
	}
	fprintf(f, "\n");
}

void ProcsDistribution::printOutMapping(FILE *f) {
	if (grpList.empty() || !mapped) return;
	vector<GroupData *>::iterator it;

	if (overallAvgOfLatencies == 0.0) {
		fprintf(f, "%5.3f:  ", mappingQuality());
	}
	else {
		//fprintf(f, "%5.3f(%d + (%5.3f+%5.3f) / %5.3f):  ", mappingQuality(), greatestGroupLatencyLevel, maxLatencyAmongGrps, avgLatencyAmongGrps, overallAvgOfLatencies);
		fprintf(f, "%5.3f(%d + %5.3f/%5.3f):  ", mappingQuality(), greatestGroupLatencyLevel, maxLatencyAmongGrps, overallAvgOfLatencies);
	}
	for (it = grpList.begin(); it != grpList.end(); it++) {
		fprintf(f, " [%d->", (*it)->getNrOfProcs());
		for (int j = 0; j < (*it)->getNrOfLoadTickets(); j++) {

			fprintf(f, "%s(%d)", (*it)->getMappingTicketByIndex(j)->getHost()->getHostname(), (*it)->getMappingTicketByIndex(j)->getNrOfProcs());
		}
		fprintf(f, "] ");
	}
	fprintf(f, "\n\n");
}

/****************************************************************************
 * Calculate Maximum Latency value among the groups
 *
 * //calculating average between each process of different groups is necessary
 * //(instead of between each hosts of different groups)
 * //because otherwise more hosts in the mapping can result lower average rate
 * //(if the max latency is the same, even if there are more slower connection in it)
 * //int nrOfProcsOnHost1 = (*git1)->getTmpTicketByIndex(i)->getNrOfProcs();
 * //int nrOfProcsOnHost2 = (*git2)->getTmpTicketByIndex(j)->getNrOfProcs();
 *
 * //avgTmpAmongGrps += nrOfProcsOnHost1*nrOfProcsOnHost2*matrix->getValue(hostIndex1, hostIndex2)->getValue();
 * //counter+= nrOfProcsOnHost1*nrOfProcsOnHost2;
 * //+ all connection within the hosts - latency is 0.0
 * //counter = counter + nrOfProcsOnHost1*(nrOfProcsOnHost1-1)/2 + nrOfProcsOnHost2*(nrOfProcsOnHost2-1)/2;
 ****************************************************************************/
void ProcsDistribution::calculateMaxLatencyAmongGrps(enum _schema_types type, LatencyMatrix *matrix) {
	maxTmpAmongGrps = 0.0;
	//int counter = 0;
	//avgTmpAmongGrps = 0.0;
	if (grpList.size()>1) { //if we have more than one group
		if (type == SCHEMA_GROUPS || edges == NULL) {
			vector<GroupData *>::iterator git1;
			for (git1 = grpList.begin(); git1 != grpList.end()-1; git1++) {
				vector<GroupData *>::iterator git2;
				for (git2 = git1+1; git2 != grpList.end();git2++) {
					for (int i=0; i < (*git1)->getNrOfTmpTickets(); i++) {
						int hostIndex1 = (*git1)->getTmpTicketByIndex(i)->getHost()->getHostId();
						for (int j=0; j < (*git2)->getNrOfTmpTickets(); j++) {
							int hostIndex2 = (*git2)->getTmpTicketByIndex(j)->getHost()->getHostId();
							if (maxTmpAmongGrps < matrix->getValue(hostIndex1, hostIndex2)->getValue()) {
								maxTmpAmongGrps = matrix->getValue(hostIndex1, hostIndex2)->getValue();
							}

						} //for j
					} //for i
				} //for git2
			}  //for git1
		} //if (type == SCHEMA_GROUPS)

		else { //in case of GRAPH, RING and TREE if edges are defined
			for (int k=0; k<getNrOfGroups(); k++) {
				for (int l = k+1; l<getNrOfGroups(); l++) {
					if (getEdge(k, l) != 0) { //only the connection with edges are counted
						GroupData *g1 = getGroupByIndex(k);
						GroupData *g2 = getGroupByIndex(l);
						for (int i=0; i < g1->getNrOfTmpTickets(); i++) {
							int hostIndex1 = g1->getTmpTicketByIndex(i)->getHost()->getHostId();
							for (int j=0; j < g2->getNrOfTmpTickets(); j++) {
								int hostIndex2 = g2->getTmpTicketByIndex(j)->getHost()->getHostId();
								if (maxTmpAmongGrps < matrix->getValue(hostIndex1, hostIndex2)->getValue()) {
									maxTmpAmongGrps = matrix->getValue(hostIndex1, hostIndex2)->getValue();
								}

							} //for j
						} //for i
					}
				} //for g2
			}  //for g1
		} //in case of GRAPH, RING and TREE if edges are defined
	} //if (grpList()>1)

	else if (grpList.size()==1) { //if we have only one group
		GroupData *grp = grpList.back();
		for (int i=0; i < grp->getNrOfTmpTickets()-1; i++) {
			int hostIndex1 = grp->getTmpTicketByIndex(i)->getHost()->getHostId();
			for (int j=i+1; j < grp->getNrOfTmpTickets(); j++) {
				int hostIndex2 = grp->getTmpTicketByIndex(j)->getHost()->getHostId();
				if (maxTmpAmongGrps < matrix->getValue(hostIndex1, hostIndex2)->getValue()) {
					maxTmpAmongGrps = matrix->getValue(hostIndex1, hostIndex2)->getValue();
				}
				//avgTmpAmongGrps += matrix->getValue(hostIndex1, hostIndex2)->getValue();
				//counter++;
			} //for j
		} //for i
	} //else
	//avgTmpAmongGrps /= counter;
}
/***************************************************************************
 * Currently applied quality function:
 *       maxGrpLatencyLevel+maxLatencyValueAmongGrps/overallAvgLatencyValue
 *
 * Possible alternate quality function:
 *       maxGrpLatencyLevel+(maxLatencyValueAmongGrps+avgLatencyValueamongGrps)/(2*overallAvgLatencyValue)
 *
 * divisor 2 is might not necessary
 * (Take into account the avg latency, too while emphesize the importance of the upper limit
 *  of the latency)
 ***************************************************************************/
bool ProcsDistribution::compareAlternativeMappings(enum _schema_types type, LatencyMatrix *matrix) {
	calculateGreatestGroupLatencyLevel();
	calculateMaxLatencyAmongGrps(type, matrix);
	//printf("Quality: %d + %5.3f / %5.3f = %5.3f\n", greatestGroupTmpLevel, maxTmpAmongGrps, overallAvgOfLatencies, tmpQuality()); //BK_TEST
	bool exchanged = false;
	if (!isMapped() || tmpQuality() < mappingQuality()) {
		setMapped();
		maxLatencyAmongGrps = maxTmpAmongGrps;
		avgLatencyAmongGrps = avgTmpAmongGrps;
		greatestGroupLatencyLevel = greatestGroupTmpLevel;
		vector<GroupData *>::iterator i;
		for (i = grpList.begin(); i != grpList.end(); i++) {
			(*i)->copyTmpTickets();
		}
		exchanged =true;
	}
	return exchanged;
}

void ProcsDistribution::resetLastCheckedLCEntries() {
	vector<GroupData *>::iterator i;
	for (i = grpList.begin(); i != grpList.end(); i++) {
		(*i)->setIndexOfLatencyCluster(0);
	}
}
