/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2010 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <float.h>

#include <vector>
#include <list>

#include <taag_basic.h>
#include <taag_errno.h>
#include <schema_parser.h>

#include "GroupData.h"
#include "HostData.h"
#include "ListOfHosts.h"
#include "ProcsDistribution.h"
#include "LatencyValue.h"
#include "LatencyMatrix.h"
#include "LatencyCluster.h"
#include "data2XMLmapping.h"

#define MAX_LATENCY_LEVEL_FACTOR 4.0
#define MIN_LATENCY_LEVEL_FACTOR 2.0
#define LATENCY_DECREMENT 0.5

using namespace std;

static int processesPerCPU = 1; //This Factor determines, how many processes cam be scheduled to one CPU
static bool taagrun = false;
static bool dumprsl = false;
static bool logfile = false;
static char logfilename[TAAG_MAX_FILENAME_STRING];
static FILE *logFileDesc;
static char xmlfilename[TAAG_MAX_FILENAME_STRING];
static enum _schema_types type;
static char executable[TAAG_MAX_FILENAME_STRING];
static char directory[TAAG_MAX_FILENAME_STRING];

static char nwsNameServer[TAAG_MAX_HOSTNAME_STRING] = "localhost";
static unsigned int nwsNameServerPort = 8090;

static int nrOfProcs;
static int nrOfGrps;
static int nrOfEdges;
static int minSizeOfGrps;
static int divisor;

static int maxSizeOfGrps = 0;

static int depth; //for SCHEMA_TREE

static int *grpSizes;
static int *edges;

static ListOfHosts *hostList;
static LatencyMatrix *latencyMatrix;

static vector<ProcsDistribution *> procsDistributions;
static vector<LatencyCluster *> latencyClusters;
static vector<int> orderOfGrpsInGraph;

double overallAvgOfLatencies = 0.0;

static int indexOfBestDistribution;
static double qualityOfBestDistribution = DBL_MAX; //A Big Value

extern int _nws_setNameServer(char *, unsigned int);
extern int _nws_getHosts(ListOfHosts *, int);
extern int _nws_getLatencies(int, LatencyMatrix *);

int initGridResources() {
	int retVal;
	_nws_setNameServer(nwsNameServer, nwsNameServerPort);
	hostList = new ListOfHosts();

	retVal = _nws_getHosts(hostList, processesPerCPU);
	if (!retVal==TAAG_SUCCESS) {
			if (logfile) fclose(logFileDesc);
			exit(retVal);
	}

	if (logfile) {
		fprintf(logFileDesc,"Hostnames and Latency Values:\n");
		hostList->printOutHosts(logFileDesc);  //BK_TEST
	}

	int nrOfHosts = hostList->getNrOfHosts();
	if (nrOfHosts==0) {
		fprintf(stderr,"Error: No hosts are available.\n");
		if (logfile) fclose(logFileDesc);
		exit(TAAG_ERR_UNKNOWN);
	}
	latencyMatrix = new LatencyMatrix(nrOfHosts);
	for (int i = 0; i < nrOfHosts; i++) {
		latencyMatrix->initValue(0.0, 0.0, i, i);
	} //for

	if (nrOfHosts>1) {
		retVal = _nws_getLatencies(hostList->getNrOfHosts(), latencyMatrix);
		if (!retVal==TAAG_SUCCESS) {
			if (logfile) fclose(logFileDesc);
			exit(retVal);
		}
	}
	if (logfile) latencyMatrix->printOutLatencyMatrix(logFileDesc); //BK_TEST
	return TAAG_SUCCESS;
}

int freeGridResources() {
	delete latencyMatrix;
	delete hostList;

	return TAAG_SUCCESS;
}

int initSchema(char *filename) {
	char s_data[TAAG_MAX_HOSTNAME_STRING];
	int retVal;

	if (_taag_schema_openFile(filename, s_data)) {
			fprintf(stderr,"The XML file \"%s\" does not exist.\n", filename);
			return(TAAG_ERR_XML);
	}
	if (_taag_schema_validate()) return(TAAG_ERR_XML);

	if (!strcmp(s_data, "groups")) type = SCHEMA_GROUPS;
	else if (!strcmp(s_data, "graph")) type = SCHEMA_GRAPH;
	else if (!strcmp(s_data, "tree")) type = SCHEMA_TREE;
	else if (!strcmp(s_data, "ring")) type = SCHEMA_RING;
	else return (TAAG_ERR_SCHEMA);

	retVal = _taag_schema_getExecutable(executable);
	if (!retVal) {
		fprintf(stderr, "Executable: %s\n", executable);
		if (logfile) fprintf(logFileDesc, "Executable: %s\n", executable);
	}
	retVal = _taag_schema_getDirectory(directory);
	if (!retVal) {
		fprintf(stderr, "Default Dir: %s\n", directory);
		if (logfile) fprintf(logFileDesc, "Default Dir: %s\n", directory);
	}

	if (logfile) fprintf(logFileDesc, "The Given Schema: ");
	switch(type) {
	case SCHEMA_GROUPS: retVal = _taag_schema_getGroups(&nrOfProcs, &minSizeOfGrps, &divisor);
						fprintf(stderr, "GROUPS { %d, %d, %d }\n", nrOfProcs, minSizeOfGrps, divisor);
						if (logfile) fprintf(logFileDesc, "GROUPS { %d, %d, %d }\n", nrOfProcs, minSizeOfGrps, divisor);
						break;
	case SCHEMA_TREE:	retVal = _taag_schema_getTree(&nrOfProcs, &depth, &minSizeOfGrps);
						divisor=(int)powl(2, depth-2);
						fprintf(stderr, "TREE { %d, %d, %d }\n", nrOfProcs, depth, minSizeOfGrps);
						if (logfile) fprintf(logFileDesc, "TREE { %d, %d, %d }\n", nrOfProcs, depth, minSizeOfGrps);
						break;
	case SCHEMA_RING:	retVal = _taag_schema_getRing(&nrOfProcs, &minSizeOfGrps, &divisor);
						fprintf(stderr, "RING { %d, %d, %d }\n", nrOfProcs, minSizeOfGrps, divisor);
						if (logfile) fprintf(logFileDesc, "RING { %d, %d, %d }\n", nrOfProcs, minSizeOfGrps, divisor);
						break;
	case SCHEMA_GRAPH:	retVal = _taag_schema_getNumberOfGroups(&nrOfGrps);
						if (retVal != TAAG_SUCCESS) return retVal;
						grpSizes = (int *)calloc(nrOfGrps,sizeof(int));
						int tmp = nrOfGrps*nrOfGrps;
						edges = (int *)calloc(tmp, sizeof(int));
						//Initializing
						for (int i = 0; i < tmp; i++) {
							if (i/nrOfGrps != i%nrOfGrps){
								*(edges+i) = -1;
							} //if
							else {
								*(edges+i) = 0;
							} //else
						} //for
						retVal = _taag_schema_getGroupSizes(nrOfGrps, grpSizes);
						if (retVal != TAAG_SUCCESS) return retVal;
						retVal = _taag_schema_getEdgeMatrix(nrOfGrps, &nrOfEdges, edges);
						fprintf(stderr, "GRAPH { %d, %d, [%d", nrOfGrps, nrOfEdges, *grpSizes);
						if (logfile) fprintf(logFileDesc, "GRAPH { %d, %d, [%d", nrOfGrps, nrOfEdges, *grpSizes);

						nrOfProcs = *grpSizes;
						for (int i=1; i<nrOfGrps; i++) {
							nrOfProcs += *(grpSizes+i);
							fprintf(stderr, ", %d", *(grpSizes+i));
							if (logfile) fprintf(logFileDesc, ", %d", *(grpSizes+i));
						}
						fprintf(stderr, "], [");
						if (logfile) fprintf(logFileDesc, "], [");
						int comma=TAAG_FALSE;
						for (int i=0; i<nrOfGrps; i++) {
							for (int j=i+1; j<nrOfGrps; j++) {
								if (*(edges+i*nrOfGrps+j)==1) {
									if (comma) {
										fprintf(stderr, ", ");
										if (logfile) fprintf(logFileDesc, ", ");
									}
									else {
										comma=TAAG_TRUE;
									}
									fprintf(stderr, "[%d, %d]", i, j);
									if (logfile) fprintf(logFileDesc, "[%d, %d]", i, j);
								} //if
							} //for
						} //for
						fprintf(stderr, "] }\n");
						if (logfile) fprintf(logFileDesc, "] }\n");

						break;
	} //switch
	return retVal;;
}

int freeSchema() {
	if (type == SCHEMA_GRAPH) {
		free(grpSizes);
		free(edges);
	}
	return TAAG_SUCCESS;
}

int cleanUpAfterMapping() {
	while (!procsDistributions.empty()) {
		ProcsDistribution *pd = procsDistributions.back();
		procsDistributions.pop_back();
		delete pd;
	}
	while (!latencyClusters.empty()) {
		LatencyCluster *pd = latencyClusters.back();
		latencyClusters.pop_back();
		delete pd;
	}
	return TAAG_SUCCESS;
}

void createEdges(ProcsDistribution *dist) {
	if (dist->getNrOfGroups()==1) return;
	switch(type) {
		case SCHEMA_GROUPS:
			return;
			break;
		case SCHEMA_RING:
			dist->initEdgeMatrix();
			for (int i=0; i<dist->getNrOfGroups()-1; i++) {
				dist->setEdge(i, i+1);
			}
			dist->setEdge(0, dist->getNrOfGroups()-1);
			break;
		case SCHEMA_TREE:
			break;
		case SCHEMA_GRAPH:
			dist->initEdgeMatrix();
			for (int i=0; i<nrOfGrps; i++) {
				for (int j=i+1; j<nrOfGrps; j++) {
					if (*(edges+i*nrOfGrps+j)==1) {
						dist->setEdge(i, j);
					} //if
				} //for
			} //for
			break;
	}
}

/********************************************
 * The evaluation is positiv
 * -either the distribution contains one group
 * -or there is at least one group house size is equal with the cpacaity of one of the latency cluster.
 ********************************************/
bool evalGroup(int nrProcsInGrp) {
	//if (nrOfProcs == nrProcsInGrp) return true;

	bool fit = false;
	vector<LatencyCluster *>::iterator it;
	for (it = latencyClusters.begin(); it!=latencyClusters.end() && !fit; it++) {
		if (nrProcsInGrp == (*it)->getProcsCapacity()) {
			fit = true;
		}
	}
	return fit;
}

/***********************************************
 * Maybe it is necessary for divisor usage
 * because sometimes no distribution
 * with positive eval is returned otherwise
 ***********************************************/
bool evalDist(int mgrs, ProcsDistribution *dist) {
	bool fit = true;
	int nrOfGrp = dist->getNrOfGroups();
	for (int i=0; i<nrOfGrp; i++) {
		int gsize = dist->getGroupByIndex(i)->getNrOfProcs();
		if ((gsize != (nrOfProcs-mgrs)/divisor)
				&&
			(gsize != (nrOfProcs-mgrs)/divisor+(nrOfProcs-mgrs)%divisor)) {
			fit = false;
		}
	}

	return fit;
}

void createDistributions(int nrProcs, int nrOfGrp, int currentGrp, ProcsDistribution *dist, bool positiveEval) {

	if (currentGrp+1 == nrOfGrp) {
		GroupData *grp = new GroupData(nrProcs);
		dist->addGroup(grp);
		if (!positiveEval) {
			positiveEval = evalGroup(grp->getNrOfProcs()) || evalDist(0, dist);
		}
		if (positiveEval) {
			ProcsDistribution *newDistribution = new ProcsDistribution();
			for (int i = 0; i<=currentGrp; i++) {
				GroupData *grp = new GroupData(dist->getGroupByIndex(i)->getNrOfProcs());
				newDistribution->addGroup(grp);
			}

			createEdges(newDistribution);

			procsDistributions.push_back(newDistribution);
			if (logfile) procsDistributions.back()->printOutDistribution(logFileDesc);  //BK_TEST
		}
		grp = dist->getLastGroup();
		dist->removeLastGroup();
		delete grp;
	} //if
	else if (currentGrp == 0) {
	    for (int i = minSizeOfGrps; i <= nrProcs/nrOfGrp; i++) {
	    	GroupData *grp = new GroupData(i);
	    	dist->addGroup(grp);
	    	if (!positiveEval) {
	    		positiveEval = evalGroup(grp->getNrOfProcs());
	    		createDistributions(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval);
	    		positiveEval = false;
	    	}
	    	else {
	    		createDistributions(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval);
	    	}

	    	grp = dist->getLastGroup();
	    	dist->removeLastGroup();
	    	delete grp;
	    } //for
	} //else if
	else {
	    int initsize = dist->getLastGroup()->getNrOfProcs();
	    for (int i = initsize; i <= nrProcs/(nrOfGrp-currentGrp); i++) {
	    	GroupData *grp = new GroupData(i);
			dist->addGroup(grp);
			if (!positiveEval) {
				positiveEval = evalGroup(grp->getNrOfProcs());
				createDistributions(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval);
				positiveEval = false;
			}
			else {
				createDistributions(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval);
			}

	        grp = dist->getLastGroup();
			dist->removeLastGroup();
			delete grp;
		}  //for
	 } //else
}

void createTree(int posPrev, int posCurr, int *structure, int level, ProcsDistribution *dist) {

	if (level == 0) {
		for (int i=1; i<=structure[1]; i++) {
			dist->setEdge(0, i);
		}
		procsDistributions.push_back(dist);
		if (logfile) procsDistributions.back()->printOutDistribution(logFileDesc);  //BK_TEST
	}
	else if (level == depth-2) {
		for (int i=0; i < structure[level]; i++) {
			dist->setEdge(posPrev-i, posCurr-i);
		}
		createTree(posCurr, posCurr-structure[level], structure, level-1, dist);
	} //else
	else {
		int avgNrOfChildren = structure[level+1]/structure[level];
		int remainder = structure[level+1]%structure[level];
		int childrenCounter = 0;
		for (int i=0; i < structure[level]; i++) {
			for (int j=0; j < avgNrOfChildren; j++) {
				dist->setEdge(posPrev-childrenCounter, posCurr-i);
				childrenCounter++;
			}
			if (remainder>0) {
				dist->setEdge(posPrev-childrenCounter, posCurr-i);
				childrenCounter++;
				remainder--;
			} //for
		} //for

		//while (childrenCounter<structure[level+1]) {
		//	dist->setEdge(posPrev-childrenCounter, posCurr-structure[level]+1);
		//	childrenCounter++;
		//}
		createTree(posCurr, posCurr-structure[level], structure, level-1, dist);
	}
}

void createLeaves(int nrProcs, int nrOfGrp, int currentGrp, ProcsDistribution *dist, bool positiveEval, int *structure) {
	if (currentGrp+1 == nrOfGrp) {
		GroupData *grp = new GroupData(nrProcs);
		dist->addGroup(grp);
		if (!positiveEval) {
			positiveEval = evalGroup(grp->getNrOfProcs());
		}

		int counter = 0;  //counting managers
		for (int k = 0; k<depth-1; k++) {
			counter += structure[k];
		}

		if (positiveEval || evalDist(counter, dist)) {
			ProcsDistribution *newDistribution = new ProcsDistribution();

			for (int j = 0; j < counter; j++) {
				newDistribution->addGroup(new GroupData(1)); //Root and mgr groups of the tree
			}
			for (int i = 0; i<=currentGrp; i++) {
				GroupData *grp = new GroupData(dist->getGroupByIndex(i)->getNrOfProcs());
				newDistribution->addGroup(grp);
			}

			newDistribution->initEdgeMatrix();
			createTree(newDistribution->getNrOfGroups()-1, newDistribution->getNrOfGroups()-1-structure[depth-1], structure, depth-2, newDistribution);
		}
		grp = dist->getLastGroup();
		dist->removeLastGroup();
		delete grp;
	} //if
	else if (currentGrp == 0) {
	    for (int i = minSizeOfGrps; i <= nrProcs/nrOfGrp; i++) {
	    	GroupData *grp = new GroupData(i);
	    	dist->addGroup(grp);
	    	if (!positiveEval) {
	    		positiveEval = evalGroup(grp->getNrOfProcs());
	    		createLeaves(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval, structure);
	    		positiveEval = false;
	    	}
	    	else {
	    		createLeaves(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval, structure);
	    	}

	    	grp = dist->getLastGroup();
	    	dist->removeLastGroup();
	    	delete grp;
	    } //for
	} //else if
	else {
	    int initsize = dist->getLastGroup()->getNrOfProcs();
	    for (int i = initsize; i <= nrProcs/(nrOfGrp-currentGrp); i++) {
	    	GroupData *grp = new GroupData(i);
			dist->addGroup(grp);
			if (!positiveEval) {
				positiveEval = evalGroup(grp->getNrOfProcs());
				createLeaves(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval, structure);
				positiveEval = false;
			}
			else {
				createLeaves(nrProcs-i, nrOfGrp, currentGrp+1, dist, positiveEval, structure);
			}

	        grp = dist->getLastGroup();
			dist->removeLastGroup();
			delete grp;
		}  //for
	 } //else
}

int createDistributionsInit() {
	int maxNrOfGrp;
	int nrOfTreeMgrs;
	int nrOfGrp;
	int structure[depth];
	int avgNrOfChildren;
	ProcsDistribution *dist = new ProcsDistribution();
	if (logfile) fprintf(logFileDesc, "Process Distributions\n");

	switch(type) {
		case SCHEMA_GROUPS:
		case SCHEMA_RING:
				maxNrOfGrp = nrOfProcs/minSizeOfGrps;
				for (nrOfGrp = divisor; nrOfGrp <= maxNrOfGrp; nrOfGrp+=divisor) {
					//all possible group sizes for each group
					createDistributions(nrOfProcs, nrOfGrp, 0, dist, false);
				} //for
				delete dist;
				break;
		case SCHEMA_TREE:
				nrOfTreeMgrs = 1; //ROOT
				structure[0]=1; //ROOT
				avgNrOfChildren = 2;
				for (int i = 1; i<depth-1; i++) {
					nrOfTreeMgrs += (int)powl(2, i);
					structure[i] = (int)powl(2, i);
				}
				nrOfGrp = (int)powl(2, depth-2);
				structure[depth-1] = nrOfGrp;

				if (depth > 2) {
					while(nrOfProcs >= nrOfTreeMgrs+nrOfGrp*minSizeOfGrps) {
						//printf("------Mgrs:%d, Grps:%d\n", nrOfTreeMgrs, nrOfGrp); //BK_TEST
						createLeaves(nrOfProcs-nrOfTreeMgrs, nrOfGrp, 0, dist, false, structure);
						nrOfGrp++;
						structure[depth-1] = nrOfGrp;
						structure[depth-2] = nrOfGrp; //nr of local mgrs
						nrOfTreeMgrs = structure[depth-2];
						for (int i = depth-3; i>0; i--) {
							if (avgNrOfChildren<structure[i+1]/structure[i]) {
								structure[i]++;
							}
							nrOfTreeMgrs += structure[i];
						}
						nrOfTreeMgrs++; //root
						if (avgNrOfChildren<structure[1]) avgNrOfChildren++;

					} //while
				} //if
				else {
					createLeaves(nrOfProcs-nrOfTreeMgrs, nrOfGrp, 0, dist, false, structure);
				} //else
				delete dist;
				break;
		case SCHEMA_GRAPH:
				for (int i=0; i < nrOfGrps; i++) {
					GroupData *grp = new GroupData(*(grpSizes+i));
					dist->addGroup(grp);
				} //for
				createEdges(dist);
				procsDistributions.push_back(dist);
				if (logfile) procsDistributions.back()->printOutDistribution(logFileDesc);  //BK_TEST
				break;
	}
	return TAAG_SUCCESS;
}

int createOrderedLatencyVector(list<LatencyValue *> *orderedLatencyVector) {
	int nrOfHosts = hostList->getNrOfHosts();
	int counter = 0;
	if (nrOfHosts>1) {
		for (int i=0; i<nrOfHosts; i++) {
			for (int j=i+1; j<nrOfHosts; j++) {
				LatencyValue *latencyValue=latencyMatrix->getValue(i, j);
				overallAvgOfLatencies += latencyValue->getValue();  //Calculate overall average
				counter++;

				if (orderedLatencyVector->empty()) {
					orderedLatencyVector->push_front(latencyValue);
				}
				else {
					bool inserted = false;
					list<LatencyValue *>::iterator it;
					for (it=orderedLatencyVector->begin(); it!=orderedLatencyVector->end() && !inserted; it++) {
						if (latencyValue->getValue() < (*it)->getValue()) {
							orderedLatencyVector->insert(it, latencyValue);
							inserted = true;
						} //if
					} //for
					if (!inserted) {
						orderedLatencyVector->push_back(latencyValue);
					} //if
				} //else
			} //for
		} //for
		overallAvgOfLatencies /= counter;
	} //if
	if (logfile) fprintf(logFileDesc, "The overall average latency value is : %6.4f\n", overallAvgOfLatencies); //BK_TEST
	return TAAG_SUCCESS;
}

/*****************************************************************
 * The limits of the latency levels are multiple of the minimal
 * latency value
 *****************************************************************/
int setLatencyLevels(list<LatencyValue *> *orderedLatencyVector) {
	double latencyLevelfactor = MAX_LATENCY_LEVEL_FACTOR;
	double currentLevelLimit = orderedLatencyVector->front()->getValue()*latencyLevelfactor;
	int currentLevel = 1;

	list<LatencyValue *>::iterator it;
	for (it=orderedLatencyVector->begin(); it!=orderedLatencyVector->end(); it++) {
		if ((*it)->getValue() < currentLevelLimit) {
			// WRONG_>(*it)->setLatencyLevel(currentLevel);
			latencyMatrix->getValue((*it)->getHostId1(), (*it)->getHostId2())->setLatencyLevel(currentLevel);
			latencyMatrix->getValue((*it)->getHostId2(), (*it)->getHostId1())->setLatencyLevel(currentLevel);
		}
		else {
			do {
				if (latencyLevelfactor > MIN_LATENCY_LEVEL_FACTOR) {
					latencyLevelfactor -= LATENCY_DECREMENT;
				}
				currentLevelLimit *=latencyLevelfactor;
				currentLevel++;
			} while ((*it)->getValue() >= currentLevelLimit);
			// WRONG_>(*it)->setLatencyLevel(currentLevel);
			latencyMatrix->getValue((*it)->getHostId1(), (*it)->getHostId2())->setLatencyLevel(currentLevel);
			latencyMatrix->getValue((*it)->getHostId2(), (*it)->getHostId1())->setLatencyLevel(currentLevel);
		}
	}

	return TAAG_SUCCESS;
}

/*******************************************************
 * Insert a latency cluster into the list according to
 * -its latency level (ascending order)
 * -its number of CPUs (descending order)
 * -its number of hosts (ascendening order)-> host with more CPUs taken first
 *******************************************************/
int sortedInsertOfLatencyClusters(LatencyCluster *lc) {
	bool inserted = false;
	vector<LatencyCluster *>::reverse_iterator rit;
	for (rit=latencyClusters.rbegin(); rit !=latencyClusters.rend() && !inserted; rit++) {
		if (lc->getLevel() > (*rit)->getLevel()) {
			latencyClusters.insert(rit.base(), lc);
			inserted = true;
		}
		else if ((*rit)->getLevel() ==  lc->getLevel() && (*rit)->getNrOfCPUs() > lc->getNrOfCPUs()) {
			latencyClusters.insert(rit.base(), lc);
			inserted = true;
		}
		else if ((*rit)->getLevel() ==  lc->getLevel()
				&&
		(*rit)->getNrOfCPUs() == lc->getNrOfCPUs()
			   &&
		(*rit)->getNrOfHosts() <= lc->getNrOfHosts()) {
			latencyClusters.insert(rit.base(), lc);
			inserted = true;

		}
	}
	if (!inserted) {
		latencyClusters.insert(rit.base(), lc);
		//latencyClusters.push_front(lc);
	}

	return TAAG_SUCCESS;
}

/**************************************************************************************
 * New assumption: The latency clusters are disjoint. No intersection!<-VERY IMPORTANT
 * MUCH FASTER EXECUTION (2.5x FASTER JUST the GENERATION OF THE LATENCY CLUSTERS)!!!
 * If a new latency cluster is created it is extended with the already
 * not contained latency clusters, whose level is less and first host is connected to
 * first two hosts of the current latency cluster on an appropriate level
 **************************************************************************************/
int createLatencyClusters() {
	int nrOfHosts = hostList->getNrOfHosts();
	if (nrOfHosts==1) {
		LatencyCluster *lc = new LatencyCluster(0);
		lc->addHost(hostList->getHostByIndex(0));
		sortedInsertOfLatencyClusters(lc);
		return TAAG_SUCCESS;
	}

	list<LatencyValue *> orderedLatencyVector;
	createOrderedLatencyVector(&orderedLatencyVector);
	setLatencyLevels(&orderedLatencyVector);

	//if (logfile) {
	//	fprintf(logFileDesc, "Ordered LatencyValues with Levels:\n");
	//	list<LatencyValue *>::iterator it2;  //BK_TEST
	//	for (it2=orderedLatencyVector.begin(); it2!=orderedLatencyVector.end(); it2++) {
	//		fprintf(logFileDesc, " %5.2f(%d)", (*it2)->getValue(), (*it2)->getLatencyLevel());
	//	}
	//	fprintf(logFileDesc, "\n");
	//}

	//Creating LatencyClusters
	//First host is inserted on level 0
	for (int i=0; i < hostList->getNrOfHosts(); i++) {
		LatencyCluster *lc = new LatencyCluster(0);
		lc->addHost(hostList->getHostByIndex(i));
		sortedInsertOfLatencyClusters(lc);
	}
	list<LatencyValue *>::iterator it;
	for (it=orderedLatencyVector.begin(); it != orderedLatencyVector.end(); it++) {
		//We check only those connection which are not contained by a latency cluster
		if (!(*it)->isOccupied()) {
			LatencyCluster *lc = new LatencyCluster((*it)->getLatencyLevel());
			HostData *host1 = hostList->getHostByIndex((*it)->getHostId1());
			HostData *host2 = hostList->getHostByIndex((*it)->getHostId2());
			lc->addCluster((LatencyCluster *)host1->getAddressOfHighestCluster());
			lc->addCluster((LatencyCluster *)host2->getAddressOfHighestCluster());
			lc->addHost(host1);
			lc->addHost(host2);

			//extend the new latency cluster
			vector<LatencyCluster *>::iterator it2;
			for (it2=latencyClusters.begin(); it2 != latencyClusters.end() && (*it2)->getLevel() < lc->getLevel(); it2++) {
				if (!(*it2)->isContained()) {
					int hostCandidateId = (*it2)->getHostByIndex(0)->getHostId();
					bool fitInToLC = true;
					for (int j=0; j < lc->getNrOfHosts() && fitInToLC; j++) {
						int hostMemberId = lc->getHostByIndex(j)->getHostId();
						int latencyLevel = latencyMatrix->getValue(hostMemberId, hostCandidateId)->getLatencyLevel();
						if (hostCandidateId == hostMemberId || lc->getLevel() < latencyLevel) {
							fitInToLC = false;
						} //if
					} //for j
					if (fitInToLC) {
						lc->addCluster(*it2);
					} //if
				} //if
			} //for it2
		    lc->collectHostsFromContainedClusters();
		    //Set the corresponding connection "occupied"
		    for (int i=0; i < lc->getNrOfHosts(); i++) {
				for (int j=i+1; j < lc->getNrOfHosts(); j++) {
					int hostMemberId1 = lc->getHostByIndex(i)->getHostId();
					int hostMemberId2 = lc->getHostByIndex(j)->getHostId();
					latencyMatrix->getValue(hostMemberId1, hostMemberId2)->setOccupied();
					latencyMatrix->getValue(hostMemberId2, hostMemberId1)->setOccupied();
				}  //for j
		    } //for i
			sortedInsertOfLatencyClusters(lc);
		} //if
	} //for it
	orderedLatencyVector.clear();

	if (logfile) {
		fprintf(logFileDesc, "Latency Clusters:\n");
		vector<LatencyCluster *>::iterator it3;
		for (it3=latencyClusters.begin(); it3 != latencyClusters.end(); it3++) {
			(*it3)->printOutLatencyCluster(logFileDesc);
		}
	}
	return TAAG_SUCCESS;
}

/************************************************************************
Important Remarks:
-Greater groups must be mapped first (GROUPS must be arranged according to
their sizes)
-A group never mapped to a latency cluster whose latency level higher than
 the previously mapped latency cluster
-A GROUP NEVER MAPPED AN EARLIER LATENCY CLUSTER IN THE LIST THAN ITS
 FIRST MAPPED LATENCY CLUSTER ON THE PREVIOUS BRANCH (BECAUSE OUR GOAL IS
 TO FIND CLOSE GROUPS BUT NOT TO EXCHANGE THE MAPPING OF GROUPS)

 More details: The case were the groups may exchange latency clusters
 are covered by another process partition, e.g.: Let assume we have four
 latency clusters with the same latency level and with 7, 5, 3, 3 available
 CPUs. And we have groups 5g 4g 3g. First we find the mapping 5g-->7 4g-->5
 and 3g-->3 We will have the following alternatives 4g3g-->7 5g-->5 (but is
 already covered by another process distribution with 7g and 5g(7g-7 and 5g-5).
 Therefore the only other alternative mapping which applies:
 When 3g--->other 3 (maybe it closer to the others)

 This may be not true in case of graph schema!!!!
*************************************************************************/
void grpMapping(ProcsDistribution *pd, int grpIndex) {
	GroupData *grp;
	if (type == SCHEMA_GRAPH) {
		grp = pd->getGroupByIndex(orderOfGrpsInGraph[grpIndex]);
	}
	else {
		grp = pd->getGroupByIndex(grpIndex);
	}
	//mapping manager groups of tree
	//it is optimized to the mapping of the children groups
	if (grp->getNrOfProcs()==1 && type == SCHEMA_TREE) {
		int childIndex;
		bool found = false;
		for (childIndex=grpIndex+1; childIndex < pd->getNrOfGroups(); childIndex++) {
			if (pd->getEdge(grpIndex, childIndex)) {
				found=true;
				break;
			}
		}

		int lcIndex;

		if (found) {
			GroupData *child = pd->getGroupByIndex(childIndex);
			int childHostIndex = child->getTmpTicketByIndex(0)->getHost()->getHostId();
			double minLatency = DBL_MAX;

			for (int unsigned j=0; j<latencyClusters.size(); j++) {
				if (!latencyClusters[j]->isChosen()) continue;
				if (latencyClusters[j]->getLevel() > 0) break; //manager assigned only hosts
				if (latencyClusters[j]->areEnoughProcsCapacity(1)) {
					int lcHostIndex = latencyClusters[j]->getHostByIndex(0)->getHostId();
					double latency = latencyMatrix->getValue(childHostIndex, lcHostIndex)->getValue();
					if (minLatency>latency) {
						minLatency = latency;
						lcIndex =  j;
					} //if
				} //if
			} //for j
		} //if found
		else { //if no child is found (special case-one leaf process in one group)
			for (int unsigned j=0; j<latencyClusters.size(); j++) {
				if (!latencyClusters[j]->isChosen()) continue;
				if (latencyClusters[j]->getLevel() > 0) break; //manager assigned only hosts
				if (latencyClusters[j]->areEnoughProcsCapacity(1)) {
					lcIndex =  j;
					break;
				} //if
			} //for j
		} //else
		latencyClusters[lcIndex]->mapGroupToHosts(grp);
		if (grpIndex == 0) {
			if (pd->compareAlternativeMappings(type, latencyMatrix)) {
				if (logfile) pd->printOutMapping(logFileDesc);
				//pd->printOutMapping(stdout); //BK_TEST
			}
		} //if
		else {
			grpMapping(pd, grpIndex-1);
		} //else
		grp->removeLastTmpTickets(grp->getNrOfProcs());

	} //if (grp->getNrOfProcs()==1 && type == SCHEMA_TREE)
	else { //mapping everything else
		int lastCheckedLC = grp->getIndexOfLatencyCluster();
		bool firstMapOnThisBranch = true;
		for (int unsigned j=lastCheckedLC; j<latencyClusters.size(); j++) {
			if (!latencyClusters[j]->isChosen()) continue;
			if (latencyClusters[j]->getLevel() > grp->getLatencyLevel()) break;
			if (latencyClusters[j]->mapGroupToHosts(grp)) {
				if (type != SCHEMA_GRAPH && firstMapOnThisBranch) {
					grp->setIndexOfLatencyCluster(j);
					firstMapOnThisBranch = false;
				} //if
				if (grpIndex == 0) {
					if (pd->compareAlternativeMappings(type, latencyMatrix)) {
						if (logfile) pd->printOutMapping(logFileDesc);
						//pd->printOutMapping(stdout); //BK_TEST
					}
				} //if
				else {
					grpMapping(pd, grpIndex-1);
				} //else
				grp->removeLastTmpTickets(grp->getNrOfProcs());
			} //if
		} //for j
	} //else if
}

int mapping() {
	if (procsDistributions.empty()) {
		fprintf(stderr,"Error: No resources are given.\n");
		if (logfile) fclose(logFileDesc);
		exit(TAAG_ERR_UNKNOWN);
	}
	if (logfile) fprintf(logFileDesc, "Mapping between Process Distribution and LatencyClusters:\n");

	//Pre-evaluation of corresponding latency clusters;Only those latency clusters (and their member
	//clusters) are taken whose capacity is enough for all process on the lowest level
	//Regio-preseting <-localization of possible region
	vector<LatencyCluster *>::iterator it;
	int latencyLevel = 0;
	for (it = latencyClusters.begin(); it!=latencyClusters.end(); it++) {
		if ((*it)->areEnoughProcsCapacity(nrOfProcs)) {
			latencyLevel = (*it)->getLevel();
			break;
		}
	}

	if (it==latencyClusters.end()) {
		fprintf(stderr,"Error: The mapping of the given schema is not possible on the current resources.\n");
				if (logfile) fclose(logFileDesc);
				exit(0);
	}

	//the groups in a graph must be sorted according to their size
	// a vector is applied for this purpose which contains the id of the groups
	if (type == SCHEMA_GRAPH) {
		ProcsDistribution *pd = procsDistributions[0];
		for (int i = 0; i < pd->getNrOfGroups(); i++) {
			orderOfGrpsInGraph.push_back(i);
		}
		bool sorted = false;
		for (int i = pd->getNrOfGroups()-1; i >= 1 && !sorted; i--) {
			sorted = true;
			for (int j = 0; j <= i-1; j++) {
				GroupData *grp1 = pd->getGroupByIndex(orderOfGrpsInGraph[j]);
				GroupData *grp2 = pd->getGroupByIndex(orderOfGrpsInGraph[j+1]);
				if (grp1->getNrOfProcs() > grp2->getNrOfProcs()) {
					int tmp = orderOfGrpsInGraph[j];
					orderOfGrpsInGraph[j] = orderOfGrpsInGraph[j+1];
					orderOfGrpsInGraph[j+1] = tmp;
					sorted = false;
				}
			}

		}
	}

	//printf("LatencyLevel:%d\n", latencyLevel); //BK_TEST
	for (it = latencyClusters.begin(); it !=latencyClusters.end() && (*it)->getLevel()<=latencyLevel; it++) {
		//maybe BUGGY part-> the chosen latency clusters may not cover the entire architecture!!!
		if ((*it)->getLevel()==latencyLevel //maybe +1 is required because a level of latency clusters may not cover all hosts<---that case reverse iterator must use
			&&
		  (*it)->areEnoughProcsCapacity(nrOfProcs)
		    &&
		  !(*it)->wasChosen()) {
			//(*it)->printOutLatencyCluster(stdout); //BK_TEST
			(*it)->setChosen(true);
			for (unsigned int i=0; i<procsDistributions.size(); i++) {
				ProcsDistribution *pd = procsDistributions[i];
				pd->resetLastCheckedLCEntries();
				grpMapping(pd, pd->getNrOfGroups()-1); //Groups are arranged in descending order
			} //for
			(*it)->setChosen(false);
		} //if
	} //for
	//comparing the best mapping of different process distribution
	if (type == SCHEMA_TREE) { //equal sign in the condition ???
		for (unsigned int i=0; i<procsDistributions.size(); i++) {
			ProcsDistribution *pd = procsDistributions[i];
			if (pd->isMapped() && qualityOfBestDistribution >= pd->mappingQuality()) {
				indexOfBestDistribution = i;
				qualityOfBestDistribution = pd->mappingQuality();
			}
		} //for
	}
	else {
		for (unsigned int i=0; i<procsDistributions.size(); i++) {
			ProcsDistribution *pd = procsDistributions[i];
			int nr = pd->getNrOfGroups();
			//-Important!!!No the equal sign in the first part of the following condition,
			//because earlier
			//chosen distributions contains less groups, implies better mapping.
			//-The second part of the condidition: if the quality and the nrOfGrps
			//is equal, then later mapping is the more balanced one.
			if (pd->isMapped()
					&&
				(qualityOfBestDistribution > pd->mappingQuality()
						||
				(qualityOfBestDistribution == pd->mappingQuality()
				&& procsDistributions[indexOfBestDistribution]->getNrOfGroups()==nr))) {
				indexOfBestDistribution = i;
				qualityOfBestDistribution = pd->mappingQuality();
			}
		} //for
	}


	fprintf(stderr, "-------------The best mapping is:-------------\n");
	procsDistributions[indexOfBestDistribution]->printOutMapping(stderr);
	if (logfile) {
		fprintf(logFileDesc, "-------------The best mapping is:-------------\n");
		procsDistributions[indexOfBestDistribution]->printOutMapping(logFileDesc);
	}
	return TAAG_SUCCESS;
}

int main(int argc, char **argv)
{
	int indexOfXMLArgument = 1;
	int retVal;
	time_t startTime = time(NULL);
	time_t endOfProcsDistributions, endOfLatencyClusters;

	if (argc == 1) {
		fprintf(stderr, "Error: Few arguments. Type \"taagrun -help\" for more details.\n");
		return(TAAG_ERR_COUNT);
	}
	if (!strcmp(argv[1],"-help")) {
		fprintf(stderr, "taag [options...] <SCHEMA file> [program arguments...]\n\n");
		fprintf(stderr, "\toptions:\n");
		fprintf(stderr, "\t\t-help          \t This help.\n");
		fprintf(stderr, "\t\t-run           \t Run executable on the grid according\n");
		fprintf(stderr, "\t\t               \t to the generated mapping.\n");
		fprintf(stderr, "\t\t-dumprsl       \t Displays only the generated RSL file\n");
		fprintf(stderr, "\t\t               \t on the standard output.\n");
		fprintf(stderr, "\t\t-ppc n         \t Specify the maximum number of scheduled\n");
		fprintf(stderr, "\t\t               \t \"Processes Per CPU\". The default is 1.\n");
		fprintf(stderr, "\t\t-log file      \t Create a logfile about the mapping.\n");
		fprintf(stderr, "\t\t-nwsns hostname\t NWS nameserver. The default is\n");
		fprintf(stderr, "\t\t               \t localhost.\n");
		fprintf(stderr, "\t\t-nwsport port  \t The port of the NWS nameserver. The\n");
		fprintf(stderr, "\t\t               \t default is 8090.\n");
		fprintf(stderr, "\n\n");
		fprintf(stderr, "\t[program arguments...] take effect iff the option \"-run\" is given.\n\n");
		return(TAAG_SUCCESS);
	}

	for (int i=1; i<argc; i++) {
		if (!strcmp(argv[i],"-run")) {
			if (i == argc-1) {
				fprintf(stderr, "Error: Few arguments, XML file name is missing.\n");
				return(TAAG_ERR_COUNT);
			}
			taagrun = true;
		}
		else if (!strcmp(argv[i],"-dumprsl")) {
			if (i == argc-1) {
				fprintf(stderr, "Error: Few arguments, XML file name is missing.\n");
				return(TAAG_ERR_COUNT);
			}
			dumprsl = true;
		}
		else if (!strcmp(argv[i],"-ppc")) {
			if (i>= argc-2) {
				fprintf(stderr, "Error: Few arguments, either processes/CPU factor or XML file name are missing.\n");
				return(TAAG_ERR_COUNT);
			}
			processesPerCPU = atoi(argv[++i]);
		}
		else if (!strcmp(argv[i],"-log")) {
			if (i>= argc-2) {
				fprintf(stderr, "Error: Few arguments, either log file name or XML file name are missing.\n");
				return(TAAG_ERR_COUNT);
			}
			strcpy(logfilename, argv[++i]);
			logFileDesc = fopen(logfilename, "w");
			logfile = true;
		}
		else if (!strcmp(argv[i],"-nwsns")) {
			if (i>= argc-2) {
				fprintf(stderr, "Error: Few arguments, either NWS name server or XML file name are not specified.\n");
				return(TAAG_ERR_COUNT);
			}
			strcpy(nwsNameServer, argv[++i]);
		}
		else if (!strcmp(argv[i],"-nwsport")) {
			if (i>= argc-2) {
				fprintf(stderr, "Error: Few arguments, either NWS port file name or XML file name are not specified.\n");
				return(TAAG_ERR_COUNT);
			}
			nwsNameServerPort= atoi(argv[++i]);
		}
		else {
			indexOfXMLArgument = i;
			break;
		}
	}  //for

	retVal = initSchema(argv[indexOfXMLArgument]);
	if (!retVal==TAAG_SUCCESS) {
		if (logfile) fclose(logFileDesc);
		exit(retVal);
	}

	if (type != SCHEMA_GRAPH){
		if (divisor == 1) {
			if (type != SCHEMA_TREE) {
				maxSizeOfGrps = nrOfProcs - minSizeOfGrps;
			}
			else { //if type == SCHEMA_TREE //BK_EXAMPLE
				if (depth < 2) {
					fprintf(stderr, "Error: Incorrect schema description.\n");
					if (logfile) fclose(logFileDesc);
					exit(TAAG_ERR_SCHEMA);
				}
				int tmpProcs = 1; //root
				for (int i = 1; i<depth-1; i++) {
					tmpProcs += (int)powl(2, i);
				}
				tmpProcs += minSizeOfGrps*(int)powl(2, depth-2); //each local manager has only one leave group
				if (nrOfProcs < tmpProcs) {
					fprintf(stderr, "Error: Incorrect schema description.\n");
					if (logfile) fclose(logFileDesc);
					exit(TAAG_ERR_SCHEMA);
				}
			}
		}
		else {  //if the divisor is greater than 1
			maxSizeOfGrps = nrOfProcs - (divisor-1)*minSizeOfGrps;

			if (maxSizeOfGrps<minSizeOfGrps) {
				fprintf(stderr, "Error: Incorrect schema description.\n");
				if (logfile) fclose(logFileDesc);
				exit(TAAG_ERR_SCHEMA);
			}
		}
	}

	initGridResources();


	createLatencyClusters();
	endOfLatencyClusters = time(NULL);
	createDistributionsInit();
	endOfProcsDistributions= time(NULL);
	mapping();
	switch(type) {
		case SCHEMA_GROUPS: _create_xml_mapping_heading(executable, "groups", directory, endOfLatencyClusters); break;
		case SCHEMA_TREE: _create_xml_mapping_heading(executable, "tree", directory, endOfLatencyClusters); break;
		case SCHEMA_RING: _create_xml_mapping_heading(executable, "ring", directory, endOfLatencyClusters); break;
		case SCHEMA_GRAPH: _create_xml_mapping_heading(executable, "graph", directory, endOfLatencyClusters); break;
	}
	_convert_mapping_to_XML(procsDistributions[indexOfBestDistribution], hostList);
	_add_latency_values(latencyMatrix);
	_save_xml_mapping(xmlfilename);
	cleanUpAfterMapping();
	freeSchema();
	freeGridResources();
	fprintf(stderr, "The generation of latency clusters took %5.1f seconds.\n", difftime(endOfLatencyClusters, startTime));
	fprintf(stderr, "The generation of process distributions took %5.1f seconds.\n", difftime(endOfProcsDistributions, endOfLatencyClusters));
	fprintf(stderr, "The entire mapping took %5.1f seconds.\n", difftime(time(NULL), startTime));
	if (logfile) {
		fprintf(logFileDesc, "The generation of latency clusters took %5.1f seconds.\n", difftime(endOfLatencyClusters, startTime));
		fprintf(logFileDesc,"The generation of process distributions took %5.1f seconds.\n", difftime(endOfProcsDistributions, endOfLatencyClusters));
		fprintf(logFileDesc, "The entire mapping took %5.1f seconds.\n", difftime(time(NULL), startTime));
	}

	if (logfile) fclose(logFileDesc);
	if (dumprsl) {
		char s_data[2*TAAG_MAX_FILENAME_STRING];
		char argument[TAAG_MAX_FILENAME_STRING];
		sprintf(s_data,"%srun -dumprsl %s", argv[0], xmlfilename);
		for (int i=indexOfXMLArgument+1; i<argc; i++) {
			sprintf(argument," %s", argv[i]);
			strcat(s_data, argument);
		}
		system(s_data);
	}
	if (taagrun) {
		char s_data[2*TAAG_MAX_FILENAME_STRING];
		char argument[TAAG_MAX_FILENAME_STRING];
		sprintf(s_data,"%srun %s", argv[0], xmlfilename);
		for (int i=indexOfXMLArgument+1; i<argc; i++) {
			sprintf(argument," %s", argv[i]);
			strcat(s_data, argument);
		}
		system(s_data);
	}
	return 0;
}
